from typing import List

LAST_EVALUATED_KEY = "LastEvaluatedKey"
FILTER_EXPRESSION = "FilterExpression"
INTERNAL_TIMESTAMP_KEY = "_timestamp"
INTERNAL_OBJECT_VERSION = "_object_version"
INTERNAL_TTL = "_ttl"
EMPTY_LIST: List[str] = []
